% -*- coding: utf-8 -*-

%   This file is part of the data publication of the point cloud time series acquired at Noordwijk
%   Copyright (c) [Python] 2021 Katharina Anders, 3DGeo Research Group, Heidelberg University, Germany 
%   and [Matlab] Sander Vos, Coastal Enginering group, Delft University of Technology, The Netherlands
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, version 3.

%   This program is distributed in the hope that it will be useful, but
%   WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
%   General Public License for more details (http://www.gnu.org/licenses/).

%   This is an example script procedure which 
%	(1) reads a point cloud from a laz file, 
%	(2) transforms the point cloud using a two matrixes from specified supplementary files, and
%	(3) writes the transformed points to an output laz file.
%	Note: Writing compressed las files (i.e. laz) requires the availability of the laszip.dll in the program environment. Otherwise, only las format can be written.

%   This script uses LASio to read and write las files
%   Matthew Parkan (2021). LASio (https://www.mathworks.com/matlabcentral/fileexchange/66033-lasio), 
%   MATLAB Central File Exchange. Retrieved March 11, 2021. 

%   This script uses lastools to uncompress laz files as other available
%   laz read tools are not able to read the additional attributes in the
%   laz file.
%   https://laszip.org/. 

%   Example: pointcloud_processing_functions('190714_010242.laz','190714_010242_trafomat.txt','Global_transformation_matrix_NW_before_2dec2020.dat','test.laz')

function pointcloud_processing_functions(infile,td_trafofile,gl_trafofile,outfile)
    %infile       : path to las file of point cloud
    %td_trafofile : path to file with time dependent transformation matrix
    %gl_trafofile : path to file with time dependent transformation matrix
    %outfile      : path to las file to store transformed point cloud

    % Check number of input arguments
    if nargin<4
        disp('Script arguments could not be parsed correctly. Check specified arguments.')
        disp('USAGE: %s <path to input point cloud> <path to file transformation matrix> <path to file to store output point cloud>')
        disp('EXAMPLE: %s 161113_000123.laz 161113_000123_trafomat.txt SOP-matrix_Kijkduin.dat 161113_000123_trafo.laz')
        error('To few input arguments.');
    end
    
    % Check availability Laszip
    if ispc
        if not(exist('laszip-cli.exe'))
            error ('Laszip not available.') 
        end     
    elseif isunix
        if not(exist('laszip-cli'))
                error ('Laszip not available.') 
        end 
    else
        disp('Platform not supported')
    end
    
    % Check LasIO
    if not(exist('lasread.m')&exist('laswrite.m'))
        error ('LASIO read/write not available.') 
    end 
    
    % Check if files exist
    if isfile(infile)==0
        error ('Specified input point cloud cannot be accessed. Please check for file: %s',infile)
    else
        li=length(infile);
        ext=infile(li-2:li);
        if not(ext=='las')|not(ext=='laz')
          error ('Input file no las/laz file.') 
        end   
    end
    if isfile(td_trafofile)==0 
        error ('Time dependent transformation matrix missing. Please check for file: %s',td_trafofile)
    end
    if isfile(gl_trafofile)==0 
        error ('Global dependent transformation matrix missing. Please check for file: %s',gl_trafofile)
    end
    
    % read input point cloud from las/laz
    [points] = las2arr(infile,true,true);
 
    % read time dependent transformation matrix and apply transformation to points
    trafo_mat_local = load(td_trafofile);
    points_transformed_local = transform_points(points, trafo_mat_local);

    % read global transformation matrix and apply transformation to points
    trafo_mat_global = load(gl_trafofile);
    points_transformed_global = transform_points(points_transformed_local, trafo_mat_global);

    % store transformed point cloud to las/laz of xyz depending on extension output file
    li=length(outfile);
    ext=outfile(li-2:li);
    if (ext=='las')|(ext=='laz')
        arr2laz(points_transformed_global, outfile);
    elseif (ext=='xyz')
        arr2xyz(points_transformed_global, outfile);
    end  
%   
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Functions%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [pointsout,lasheader]=las2arr(pointcloudfile, get_intensity, filter_echos)
%Function which reads and returns point data from las/laz point clouds.
%The function also output the las header needed for writing the results
%to las/laz.
%param pointcloudfile: path to input las/laz point cloud
%return: 2D array of 3D points with shape (N,3), optionally with 
%additional 1D array of intensity information and las header

% open the las/laz file and read out point data into numpy array of shape (N,3)
    
    % In case laz file, unzip file with laszip-cli
    pl=length(pointcloudfile);
    if pointcloudfile(pl-2:pl) == 'laz'  
        if ispc
             commandstring = ['laszip-cli.exe ' pointcloudfile];
        else
             commandstring = ['laszip-cli ' pointcloudfile];
        end
        status = system(commandstring);
        disp(append('Laz conversion:',commandstring))
        pointcloudfile = strrep(pointcloudfile,'.laz','.las');   
    end
    
    % open the las/laz file and read out point data into array 
    inFile=LASread(pointcloudfile);
    points=[inFile.record.x inFile.record.y inFile.record.z];
    
    % debug
    if 1==0
        pc=inFile;
        disp(sprintf('X_min,max=%0.5f,%0.5f',min(pc.record.x), max(pc.record.x)));
        disp(sprintf('Y_min,max=%0.5f,%0.5f',min(pc.record.y), max(pc.record.y)));
        disp(sprintf('Z_min,max=%0.5f,%0.5f',min(pc.record.z), max(pc.record.z)));
        disp(sprintf('Phi_min,max=%0.5f,%0.5f',min(pc.record.phi), max(pc.record.phi)));
        disp(sprintf('Theta_min,max=%0.5f,%0.5f',min(pc.record.theta), max(pc.record.theta)));
        disp(sprintf('Time_min,max=%0.5f,%0.5f',min(pc.record.time), max(pc.record.time)));
        disp(sprintf('Range_min,max=%0.5f,%0.5f',min(pc.record.range), max(pc.record.range)));
        disp(sprintf('Reflectance_min,max=%0.5f,%0.5f',min(pc.record.reflectance), max(pc.record.reflectance)));
        disp(sprintf('Amplitude_min,max=%0.5f,%0.5f',min(pc.record.amplitude), max(pc.record.amplitude)));
        disp(sprintf('Deviation_min,max=%0.5f,%0.5f',min(pc.record.deviation), max(pc.record.deviation)));
        disp(sprintf('Echo_min,max=%0.5f,%0.5f',min(pc.record.echo), max(pc.record.echo)));
    end
    if 0==1
        figure;plot(inFile.record.echo);title('echo')
        figure;plot(inFile.record.reflectance);title('eflectance')
        figure;plot(inFile.record.amplitude);title('amplitude')
        figure;plot(inFile.record.deviation);title('deviation')
        figure;plot(inFile.record.range);title('range')
        figure;plot(inFile.record.phi);title('phi')
        figure;plot(inFile.record.theta);title('theta')
        figure;plot(inFile.record.time);title('time')
    end
    
    % read out intensity information from the reflectance attribute
	
    if get_intensity == true
        if isfield(inFile.record,'reflectance')
          reflectances = inFile.record.reflectance;
          reflectances(reflectances>100)=100;
        end
    end 

    
    % filter point cloud by echo types single (0) and last (3), removing first and intermediate returns
    if filter_echos == true
        if isfield(inFile.record,'echo')
            echotypes = inFile.record.echo;
            sel_idxs = find(echotypes == 0 | echotypes == 3);
            if get_intensity == true
                reflectances_filtered = reflectances(sel_idxs);
                reflectances = reflectances_filtered;
            end
            points_filtered = points(sel_idxs,:);
            points = points_filtered;
        end
    end
    
    if get_intensity==false
		pointsout = points;
    else
        if isfield(inFile.record,'reflectance')
            pointsout = [points reflectances];
        else
            pointsout = [points];
        end
        
    end
end

function pts_trafo=transform_points(points, trafomat)
%	Applies a rigid transformation, i.e. rotation and translation, to 3D point data.
%	:param points: 2D array of 3D points with shape (N,3)
%	:param trafomat: 2D array of rigid transformation matrix with 3x3 rotation and 1x3 translation parameters
%	:return: transformed points in 2D array of shape (N,3)
    rotation = trafomat(1:3,1:3);
	translation = trafomat(1:3,4);
% check size array
    nn=size(points);
    if nn(1,2)>3
        reflectances = points(:,4);
    end 
	pts_rot = points(:,1:3)*rotation';
	pts_translate = pts_rot + translation';
    if nn(1,2)>3
        pts_trafo = [pts_translate reflectances];
    else
        pts_trafo = pts_translate;
    end
end

function arr2laz(outpoints, outfilepath)
%	Writes an input array of 3D points to a specified output las/laz file.
%   The Infile is used to create the las header
%	:param outpoints: 2D array of shape (N,3) that should be written to file
%   :param in
%	:param outfilepath: path to output las/laz point cloud (existent files will be overwritten)
%	:return: nothing

%   prepare variables
    nn=size(outpoints);
    lengtharray=nn(1,1);
    array_0=zeros(lengtharray,1);
    array_1=array_0+1;
    array_logical=(array_1==0);
    array_int8=zeros(lengtharray,1,'int8');
    array_uint8=zeros(lengtharray,1,'uint8');
    array_uint16=zeros(lengtharray,1,'uint16');    

%   load LAS header
    lasheader=LASread('lasheader.las');
    outFile=lasheader;

% set header    
    outFile.header.system_identifier = 'pointcloud_processing_functions.m';
    outFile.header.generating_software = 'Matlab' 
    
    outFile.header.n_point_records = length(outpoints(:,1));
    outFile.header.n_point_records_extended = length(outpoints(:,1));
% set max/min values
    outFile.header.min_x = min(outpoints(:,1));
    outFile.header.max_x = max(outpoints(:,1));    
    outFile.header.min_y = min(outpoints(:,2));
    outFile.header.max_y = max(outpoints(:,2));
    outFile.header.min_z = min(outpoints(:,3));
    outFile.header.max_z = max(outpoints(:,3));   
% set values
    outFile.record.x = outpoints(:,1);
    outFile.record.y = outpoints(:,2); 
    outFile.record.z = outpoints(:,3);
    outFile.record.intensity = array_uint16;
    outFile.record.return_number = array_uint8;
    outFile.record.number_of_returns = array_uint8;
    outFile.record.scan_direction_flag = array_logical;
    outFile.record.flightline_edge_flag = array_logical;
    outFile.record.classification = array_uint8;
    outFile.record.classification_synthetic = array_logical;
    outFile.record.classification_keypoint = array_logical;
    outFile.record.classification_withheld = array_logical;
    outFile.record.scan_angle = array_int8;
    outFile.record.user_data = array_uint8;
    outFile.record.point_source_id = array_uint16;
    outFile.record.range = array_0;
    outFile.record.theta = array_0;
    outFile.record.phi = array_0;
    if nn(1,2)>3
        outFile.record.reflectance = outpoints(:,4);
    else
        outFile.record.reflectance = array_0;
    end
    outFile.record.amplitude = array_0;
    outFile.record.deviation = array_uint16;
    outFile.record.echo = array_uint8;
    outFile.record.time = array_0;
 
    % write file
    

    % In case laz file, unzip file with laszip-cli
    pl=length(outfilepath);
    if outfilepath(pl-2:pl) == 'laz'  
        outfilepathlas = strrep(outfilepath,'.laz','.las'); 
        LASwrite(outFile, outfilepathlas , 'version', 14);   
        if ispc
	    commandstring = ['laszip-cli.exe -i ' outfilepathlas ' -o ' outfilepath];
	else
	    commandstring = ['laszip-cli -i ' outfilepathlas ' -o ' outfilepath];
        end       
        status = system(commandstring);
        disp(append('Laz conversion:',commandstring))
 
    else
        LASwrite(outFile, outfilepath , 'version', 14);
    end

end

function arr2xyz(outpoints, outfilepath)
%	Writes an input array of 3D points to a specified output las/laz file.
%   The Infile is used to create the las header
%	:param outpoints: 2D array of shape (N,3) that should be written to file
%   :param in
%	:param outfilepath: path to output las/laz point cloud (existent files will be overwritten)
%	:return: nothing

%save(outfilepath,'outpoints','-ascii');
dlmwrite(outfilepath,outpoints,'precision','%.2f')

end