# -*- coding: utf-8 -*-

# This file is part of the data publication of the point cloud time series acquired at Noordwijk
# Copyright (c) [Python] 2021 Katharina Anders, 3DGeo Research Group, Heidelberg University, Germany
# and [Matlab] Sander Vos, Coastal Enginering group, Delft University of Techonology, The Netherlands.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details (http://www.gnu.org/licenses/).

#   This is an example script procedure which 
#	(1) reads a point cloud from a laz file, 
#	(2) transforms the point cloud using a two matrixes from specified supplementary files, and
#	(3) writes the transformed points to an output laz file.
#	Note: Writing compressed las files (i.e. laz) requires the availability of the laspy in the program environment. 

#   This script uses laspy to read and write las files
#   Laspy: https://pypi.org/project/laspy

import sys
try:
	import numpy as np
except:
	print("Could not import required module: numpy")
	print("Please install the module for usage of the script functions")
	sys.exit()

try:
	import laspy
except:
	print("Could not import required module: laspy")
	print("Please install the module for usage of the script functions")
	sys.exit()

def las2arr(pointcloudfile, get_intensity=False, filter_echos=False):
	'''
	Function which reads and returns point data from las/laz point clouds.
	:param pointcloudfile: path to input las/laz point cloud
	:param get_intensity: (optional) set True if intensity (reflectance values) should be returned with point data
	:param filter_echos: (optional) set True if only points from single and last echoes should be returned
	:return: 2D array of 3D points with shape (N,3), optionally with additional 1D array of intensity information
	'''

	# open the las/laz file and read out point data into numpy array of shape (N,3)
	inFile = laspy.file.File(pointcloudfile, mode="r")
	points = np.vstack((inFile.x, inFile.y, inFile.z)).transpose()

	# read out intensity information from the reflectance attribute
	if get_intensity == True:
		reflectances = inFile.reflectance # reads the raw values
		reflectances_scale = inFile.header.vlrs[0].reflectance.scale[0] # access the scale of attribute values
		if reflectances_scale < 1.0:
			reflectances = reflectances.astype(float) * reflectances_scale

	# filter point cloud by echo types single (0) and last (3), removing first and intermediate returns
	if filter_echos == True:
		echotypes = inFile.echo
		sel_idxs = np.where((echotypes == 0) | (echotypes == 3))
		if get_intensity==True:
			reflectances_filtered=reflectances[sel_idxs]
			reflectances = reflectances_filtered
		points_filtered=points[sel_idxs]
		points = points_filtered

	if get_intensity==False:
		return (points)
	else:
		return (points, reflectances)

def arr2xyz(outpoints, outfilepath):
	'''
	Writes an input array of 3D points to a specified output xyz file.
	:param outpoints: 2D array of shape (N,3) that should be written to file
	:param outfilepath: path to output las/laz point cloud (existent files will be overwritten)
	:return: nothing
	'''
	with open(outfilepath,"w") as xyz_file:
		for i in range(0,len(outpoints[:, 0])):
			xyz_file.write("{:10.2f} {:10.2f} {:6.2f} {:6.2f}\n".format(outpoints[i, 0],outpoints[i, 1],outpoints[i, 2],outpoints[i, 3]))
	return

def arr2laz(outpoints, outfilepath):
	'''
	Writes an input array of 3D points to a specified output las/laz file.
	:param outpoints: 2D array of shape (N,3) that should be written to file
	:param outfilepath: path to output las/laz point cloud (existent files will be overwritten)
	:return: nothing
	'''
	from laspy.header import Header as LasHeader
	hdr = LasHeader(version_major=1, version_minor=4)
	outFile = laspy.file.File(outfilepath, mode="w",header=hdr)
	xmin, ymin, zmin = np.min(outpoints, axis=0)
	outFile.header.offset = [xmin, ymin, zmin]
	outFile.header.scale = [0.001, 0.001, 0.001]
	outFile.x = outpoints[:, 0]
	outFile.y = outpoints[:, 1]
	outFile.z = outpoints[:, 2]
	outFile.reflectance = outpoints[:,3]
	outFile.close()
	return

def transform_points(points, trafomat):
	'''
	Applies a rigid transformation, i.e. rotation and translation, to 3D point data.
	:param points: 2D array of 3D points with shape (N,3)
	:param trafomat: 2D array of rigid transformation matrix with 3x3 rotation and 1x3 translation parameters
	:return: transformed points in 2D array of shape (N,3)
	'''
	rotation = np.array(trafomat)[:, :3]
	translation = np.array(trafomat)[:, 3].flatten()
	pts_rot = points.dot(rotation.T)
	pts_trafo = pts_rot + translation
	return pts_trafo[:,:3]

def conversion(infile,td_trafofile,gl_trafofile,outfile):
	import os
	import numpy as np

	# read input point cloud from las/laz
	points, refs = las2arr(infile,get_intensity=True,filter_echos=True)

	# read time dependent transformation matrix and apply transformation to points
	trafo_mat_local = np.loadtxt(td_trafofile)
	#trafo_mat_local = np.transpose(trafo_mat_local1)
	points_transformed_local = transform_points(points, trafo_mat_local)

	# read global transformation matrix and apply transformation to points
	trafo_mat_global = np.loadtxt(gl_trafofile)
	points_transformed_global = transform_points(points_transformed_local, trafo_mat_global)

	# add reflections to array
	refs=refs.reshape(len(refs),1)
	pointsall = np.hstack((points_transformed_global,refs))

	# store transformed point cloud to las/laz or xyz depending on the extension
	ext=outfile.strip().split("."):
	if (ext[-1] == "laz") or (ext[-1] == "las"):
		arr2laz(pointsall, outfile)
	elif (ext[-1] == "xyz"):
		arr2xyz(pointsall, outfile)
	return


if __name__ == "__main__":

	'''
	This is an example script procedure which
		(1) reads a point cloud from a laz file,
		(2) transforms the point cloud using a matrix from a specified supplementary file, and
		(3) writes the transformed points to an output laz file.
	Note: Writing compressed las files (i.e. laz) requires the availability of the laszip.dll in the program environment. Otherwise, only las format can be written.
	'''

	import os
	import numpy as np
	try:
		infile = sys.argv[1] # path to las file of point cloud
		td_trafofile = sys.argv[2] # path to file with the time dependent transformation matrix
		gl_trafofile = sys.argv[3] # path to file with the global transformation matrix
		outfile = sys.argv[4] # path to las file to store transformed point cloud
	except:
		print('Script arguments could not be parsed correctly. Check specified arguments.')
		print('USAGE: %s <path to input point cloud> <path to file transformation matrix> <path to file to store output point cloud>' % sys.argv[0])
		print('EXAMPLE: %s 190802_160246.laz 190802_160246_trafomat.txt Global_transformation_matrix_NW_before_2dec2020.dat 190802_160246_trafo.laz' % sys.argv[0])
		sys.exit()

	if not os.path.isfile(infile):
		print("Specified input point cloud cannot be accessed. Please check for file: %s" % infile)
		sys.exit()

	# read input point cloud from las/laz
	points, refs = las2arr(infile,get_intensity=True,filter_echos=True)

	# read time dependent transformation matrix and apply transformation to points
	trafo_mat_local = np.loadtxt(td_trafofile)
	points_transformed_local = transform_points(points, trafo_mat_local)

	# read global transformation matrix and apply transformation to points
	trafo_mat_global = np.loadtxt(gl_trafofile)
	points_transformed_global = transform_points(points_transformed_local, trafo_mat_global)

	# add reflections to array
	pointsall = np.hstack(points_transformed_global,refs)

	# store transformed point cloud to las/laz or xyz depending on the extension
	ext=outfile.strip().split("."):
	if (ext[-1] == "laz") or (ext[-1] == "las"):
		arr2laz(pointsall, outfile)
	elif (ext[-1] == "xyz"):
		arr2xyz(pointsall, outfile)

