#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 19 13:30:51 2022

@author: mkuschnerus
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from os import path
import laspy

def read_file(filename, after_2dec = True, read_refl=False):
    """read file if _incl.asc file is in the same folder, the alignment will be done with the angles provided in the file
       return data frame with only x,y,z """
       
    z_scanner = 55.755   # scanner elevation in m 
            
    if '.laz' in filename:
       
        inFile = laspy.read(filename)
        
        data = pd.DataFrame({'x' : np.array(inFile.x), 'y' : np.array(inFile.y), 'z' : np.array(inFile.z)})
    
        if read_refl:
            reflectance = inFile.points['reflectance']
            data['ref'] = reflectance
            
    else:
        raise ValueError('unknown file format:'+filename)
    
   
    roll = np.nan
    pitch = np.nan
        
    if not 'aligned' in filename:
    # rotate data to correct for inclination:
        
        inc_file = filename[:-4]+'_incl.asc'  
        if path.isfile(inc_file):
            temp_data = pd.read_csv(inc_file, delim_whitespace=True, header = None)
            temp_roll = temp_data.iloc[:,0].values[10:-10]  # skip first and last 10 entries because they can contain outliers
            temp_pitch = temp_data.iloc[:,1].values[10:-10] # skip first and last 10 entries because they can contain outliers
            
            roll  = np.round(temp_roll.mean(),4)
            pitch = np.round(temp_pitch.mean(),4)
            
            roll = -roll * np.pi/180.0
            pitch = - pitch *np.pi/180.0
            
            trafo_matrix = np.asarray([[np.cos(pitch), 0, np.sin(pitch)], [np.sin(pitch)*np.sin(roll), np.cos(roll), -np.cos(pitch)*np.sin(roll)],[-np.sin(pitch)*np.cos(roll), np.sin(roll), np.cos(pitch)*np.cos(roll)]])
            
            rotation = lambda inp: np.dot(np.asarray(inp), trafo_matrix)
            data[['x', 'y','z']] = data[['x', 'y','z']].apply(rotation, axis=1, result_type='expand')
                

        else: 
            print('Rotation file not found!')
   
    # correct for scanner height, so that z-coordinate corresponds to NAP and is not negative
    data.z = data.z+z_scanner
    
    #transformation for after 2nd December 2020
    if after_2dec:
        data.x = data.x+0.05
        data.y = data.y+3.09
    
    print(str(len(data))+' data points have been extracted from '+filename)
    
    return data



if __name__ == "__main__":
    
    ### example how to use it: 
    ### files need to be in folder path and _incl.asc files need to be present in the same folder in order to do the alignment.
    
    filename = 'path/210101/210101_120139.laz' 
    
    data = read_file(filename, after_2dec=True)
    
    ### plot overview of data
    plt.figure(figsize=(4,8))
    scat = plt.scatter(data.x, data.y, c=data.z.values, s=1)
    plt.xlim([-350,-145])
    plt.ylim([-250,250])
    plt.colorbar(scat)
    
    ### save to .asc file
    output_path = 'path/210101/' 
    data.to_csv(path.join(output_path,filename[-17:-4]+'_aligned.asc'), index=False, header=True)
    